@@ -5,7 +5,6 @@
 # SPDX-License-Identifier: GPL-2.0-or-later
 
 import binascii
-import re
 import sys
 
 from bitstring import BitArray, BitStream, CreationError
@@ -13,6 +12,7 @@ from bitstring import BitArray, BitStream, CreationError
 import esptool
 
 from . import util
+from typing import List
 
 
 class CheckArgValue(object):
@@ -59,8 +59,9 @@ class CheckArgValue(object):
             elif efuse.efuse_type.startswith("bytes"):
                 if new_value is None:
                     raise esptool.FatalError(
-                        "New value required for efuse '{}' "
-                        "(given None)".format(efuse.name)
+                        "New value required for efuse '{}' (given None)".format(
+                            efuse.name
+                        )
                     )
                 if len(new_value) * 8 != efuse.bitarray.len:
                     raise esptool.FatalError(
@@ -85,21 +86,34 @@ class CheckArgValue(object):
 class EfuseProtectBase(object):
     # This class is used by EfuseBlockBase and EfuseFieldBase
 
-    def get_read_disable_mask(self):
+    def get_read_disable_mask(self, blk_part=None):
+        """Returns mask of read protection bits
+        blk_part:
+            - None: Calculate mask for all read protection bits.
+            - a number: Calculate mask only for specific item in read protection list.
+        """
         mask = 0
         if isinstance(self.read_disable_bit, list):
-            for i in self.read_disable_bit:
-                mask |= 1 << i
+            if blk_part is None:
+                for i in self.read_disable_bit:
+                    mask |= 1 << i
+            else:
+                mask |= 1 << self.read_disable_bit[blk_part]
         else:
             mask = 1 << self.read_disable_bit
         return mask
 
-    def is_readable(self):
+    def get_count_read_disable_bits(self):
+        """Returns the number of read protection bits used by the field"""
+        # On the C2 chip, BLOCK_KEY0 has two read protection bits [0, 1].
+        return bin(self.get_read_disable_mask()).count("1")
+
+    def is_readable(self, blk_part=None):
         """Return true if the efuse is readable by software"""
         num_bit = self.read_disable_bit
         if num_bit is None:
             return True  # read cannot be disabled
-        return (self.parent["RD_DIS"].get() & (self.get_read_disable_mask())) == 0
+        return (self.parent["RD_DIS"].get() & self.get_read_disable_mask(blk_part)) == 0
 
     def disable_read(self):
         num_bit = self.read_disable_bit
@@ -214,13 +228,14 @@ class EfuseBlockBase(EfuseProtectBase):
 
         return [self.parent.read_reg(offs) for offs in get_offsets(self)]
 
-    def read(self):
+    def read(self, print_info=True):
         words = self.get_words()
         data = BitArray()
         for word in reversed(words):
             data.append("uint:32=%d" % word)
         self.bitarray.overwrite(data, pos=0)
-        self.print_block(self.bitarray, "read_regs")
+        if print_info:
+            self.print_block(self.bitarray, "read_regs")
 
     def print_block(self, bit_string, comment, debug=False):
         if self.parent.debug or debug:
@@ -304,7 +319,7 @@ class EfuseBlockBase(EfuseProtectBase):
                                     if rd_chunk == wr_chunk:
                                         print(
                                             "wr_chunk == rd_chunk. "
-                                            "Countinue with empty chunk."
+                                            "Continue with empty chunk."
                                         )
                                         wr_data[i : i + 6 * 8 :].set(0)
                                     else:
@@ -372,6 +387,18 @@ class EfuseBlockBase(EfuseProtectBase):
                     )
                     break
             if not self.fail and self.num_errors == 0:
+                self.read(print_info=False)
+                if self.wr_bitarray & self.bitarray != self.wr_bitarray:
+                    # if the required bits are not set then we need to re-burn it again.
+                    if burns < 2:
+                        print(
+                            f"\nRepeat burning BLOCK{self.id} (#{burns + 2}) because not all bits were set"
+                        )
+                        continue
+                    else:
+                        print(
+                            f"\nAfter {burns + 1} attempts, the required data was not set to BLOCK{self.id}"
+                        )
                 break
 
     def burn(self):
@@ -428,11 +455,12 @@ class EspEfusesBase(object):
     """
 
     _esp = None
-    blocks = []
-    efuses = []
+    blocks: List[EfuseBlockBase] = []
+    efuses: List = []
     coding_scheme = None
     force_write_always = None
     batch_mode_cnt = 0
+    postpone = False
 
     def __iter__(self):
         return self.efuses.__iter__()
@@ -477,6 +505,59 @@ class EspEfusesBase(object):
         for efuse in self.efuses:
             efuse.update(self.blocks[efuse.block].bitarray)
 
+    def postpone_efuses_from_block0_to_burn(self, block):
+        postpone_efuses = {}
+
+        if block.id != 0:
+            return postpone_efuses
+
+        # We need to check this list of efuses. If we are going to burn an efuse
+        # from this list, then we need to split the burn operation into two
+        # steps. The first step involves burning efuses not in this list. In
+        # case of an error during this step, we can recover by burning the
+        # efuses from this list at the very end. This approach provides the
+        # ability to recover efuses if an error occurs during the initial burn
+        # operation.
+
+        # List the efuses here that must be burned at the very end, such as read
+        # and write protection fields, as well as efuses that disable
+        # communication with the espefuse tool.
+        efuses_list = ["WR_DIS", "RD_DIS"]
+        if self._esp.CHIP_NAME == "ESP32":
+            # Efuses below disables communication with the espefuse tool.
+            efuses_list.append("UART_DOWNLOAD_DIS")
+            # other efuses that are better to burn at the very end.
+            efuses_list.append("ABS_DONE_1")
+            efuses_list.append("FLASH_CRYPT_CNT")
+        else:
+            # Efuses below disables communication with the espefuse tool.
+            efuses_list.append("ENABLE_SECURITY_DOWNLOAD")
+            efuses_list.append("DIS_DOWNLOAD_MODE")
+            # other efuses that are better to burn at the very end.
+            efuses_list.append("SPI_BOOT_CRYPT_CNT")
+            efuses_list.append("SECURE_BOOT_EN")
+
+        def get_raw_value_from_write(self, efuse_name):
+            return self[efuse_name].get_bitstring(from_read=False)
+
+        for efuse_name in efuses_list:
+            postpone_efuses[efuse_name] = get_raw_value_from_write(self, efuse_name)
+
+        if any(value != 0 for value in postpone_efuses.values()):
+            if self.debug:
+                print("These BLOCK0 efuses will be burned later at the very end:")
+                print(postpone_efuses)
+            # exclude these efuses from the first burn (postpone them till the end).
+            for key_name in postpone_efuses.keys():
+                self[key_name].reset()
+        return postpone_efuses
+
+    def recover_postponed_efuses_from_block0_to_burn(self, postpone_efuses):
+        if any(value != 0 for value in postpone_efuses.values()):
+            print("Burn postponed efuses from BLOCK0.")
+            for key_name in postpone_efuses.keys():
+                self[key_name].save(postpone_efuses[key_name])
+
     def burn_all(self, check_batch_mode=False):
         if check_batch_mode:
             if self.batch_mode_cnt != 0:
@@ -496,18 +577,44 @@ class EspEfusesBase(object):
                 have_wr_data_for_burn = True
         if not have_wr_data_for_burn:
             print("Nothing to burn, see messages above.")
-            return
+            return True
         EspEfusesBase.confirm("", self.do_not_confirm)
 
-        # Burn from BLKn -> BLK0. Because BLK0 can set rd or/and wr protection bits.
-        for block in reversed(self.blocks):
+        def burn_block(block, postponed_efuses):
             old_fail = block.fail
             old_num_errors = block.num_errors
             block.burn()
             if (block.fail and old_fail != block.fail) or (
                 block.num_errors and block.num_errors > old_num_errors
             ):
+                if postponed_efuses:
+                    print("The postponed efuses were not burned due to an error.")
+                    print("\t1. Try to fix a coding error by this cmd:")
+                    print("\t   'espefuse.py check_error --recovery'")
+                    command_string = " ".join(
+                        f"{key} {value}"
+                        for key, value in postponed_efuses.items()
+                        if value.any(True)
+                    )
+                    print("\t2. Then run the cmd to burn all postponed efuses:")
+                    print(f"\t   'espefuse.py burn_efuse {command_string}'")
+
                 raise esptool.FatalError("Error(s) were detected in eFuses")
+
+        # Burn from BLKn -> BLK0. Because BLK0 can set rd or/and wr protection bits.
+        for block in reversed(self.blocks):
+            postponed_efuses = (
+                self.postpone_efuses_from_block0_to_burn(block)
+                if self.postpone
+                else None
+            )
+
+            burn_block(block, postponed_efuses)
+
+            if postponed_efuses:
+                self.recover_postponed_efuses_from_block0_to_burn(postponed_efuses)
+                burn_block(block, postponed_efuses)
+
         print("Reading updated efuses...")
         self.read_coding_scheme()
         self.read_blocks()
@@ -538,6 +645,50 @@ class EspEfusesBase(object):
         else:
             raise esptool.FatalError(error_msg)
 
+    def get_block_errors(self, block_num):
+        """Returns (error count, failure boolean flag)"""
+        return self.blocks[block_num].num_errors, self.blocks[block_num].fail
+
+    def is_efuses_incompatible_for_burn(self):
+        # Overwrite this function for a specific target if you want to check if a certain eFuse(s) can be burned.
+        return False
+
+    def get_major_chip_version(self):
+        try:
+            return self["WAFER_VERSION_MAJOR"].get()
+        except KeyError:
+            return 0
+
+    def get_minor_chip_version(self):
+        try:
+            return self["WAFER_VERSION_MINOR"].get()
+        except KeyError:
+            return 0
+
+    def get_chip_version(self):
+        return self.get_major_chip_version() * 100 + self.get_minor_chip_version()
+
+    def get_major_block_version(self):
+        try:
+            return self["BLK_VERSION_MAJOR"].get()
+        except KeyError:
+            return 0
+
+    def get_minor_block_version(self):
+        try:
+            return self["BLK_VERSION_MINOR"].get()
+        except KeyError:
+            return 0
+
+    def get_block_version(self):
+        return self.get_major_block_version() * 100 + self.get_minor_block_version()
+
+    def get_pkg_version(self):
+        try:
+            return self["PKG_VERSION"].get()
+        except KeyError:
+            return 0
+
 
 class EfuseFieldBase(EfuseProtectBase):
     def __init__(self, parent, param):
@@ -553,19 +704,17 @@ class EfuseFieldBase(EfuseProtectBase):
         self.efuse_type = param.type
         self.description = param.description
         self.dict_value = param.dictionary
+        self.bit_len = param.bit_len
+        self.alt_names = param.alt_names
         self.fail = False
         self.num_errors = 0
-        if self.efuse_type.startswith("bool"):
-            field_len = 1
-        else:
-            field_len = int(re.search(r"\d+", self.efuse_type).group())
-            if self.efuse_type.startswith("bytes"):
-                field_len *= 8
-        self.bitarray = BitStream(field_len)
-        self.bit_len = field_len
+        self.bitarray = BitStream(self.bit_len)
         self.bitarray.set(0)
         self.update(self.parent.blocks[self.block].bitarray)
 
+    def is_field_calculated(self):
+        return self.word is None or self.pos is None
+
     def check_format(self, new_value_str):
         if new_value_str is None:
             return new_value_str
@@ -614,6 +763,10 @@ class EfuseFieldBase(EfuseProtectBase):
 
     def check_new_value(self, bitarray_new_value):
         bitarray_old_value = self.get_bitstring() | self.get_bitstring(from_read=False)
+
+        if not bitarray_new_value.any(True) and not bitarray_old_value.any(True):
+            return
+
         if bitarray_new_value.len != bitarray_old_value.len:
             raise esptool.FatalError(
                 "For {} efuse, the length of the new value is wrong, "
@@ -621,7 +774,10 @@ class EfuseFieldBase(EfuseProtectBase):
                     self.name, bitarray_old_value.len, bitarray_new_value.len
                 )
             )
-        if bitarray_new_value == bitarray_old_value:
+        if (
+            bitarray_new_value == bitarray_old_value
+            or bitarray_new_value & self.get_bitstring() == bitarray_new_value
+        ):
             error_msg = "\tThe same value for {} ".format(self.name)
             error_msg += "is already burned. Do not change the efuse."
             print(error_msg)
@@ -634,7 +790,7 @@ class EfuseFieldBase(EfuseProtectBase):
         else:
             if self.name not in ["WR_DIS", "RD_DIS"]:
                 # WR_DIS, RD_DIS fields can have already set bits.
-                # Do not neeed to check below condition for them.
+                # Do not need to check below condition for them.
                 if bitarray_new_value | bitarray_old_value != bitarray_new_value:
                     error_msg = "\tNew value contains some bits that cannot be cleared "
                     error_msg += "(value will be {})".format(
@@ -658,8 +814,10 @@ class EfuseFieldBase(EfuseProtectBase):
         self.save_to_block(bitarray_field)
 
     def update(self, bit_array_block):
-        if self.word is None or self.pos is None:
-            self.bitarray.overwrite(self.convert_to_bitstring(self.get()), pos=0)
+        if self.is_field_calculated():
+            self.bitarray.overwrite(
+                self.convert_to_bitstring(self.check_format(self.get())), pos=0
+            )
             return
         field_len = self.bitarray.len
         bit_array_block.pos = bit_array_block.length - (
@@ -721,3 +879,29 @@ class EfuseFieldBase(EfuseProtectBase):
         # Burn a efuse. Added for compatibility reason.
         self.save(new_value)
         self.parent.burn_all()
+
+    def get_info(self):
+        output = f"{self.name} (BLOCK{self.block})"
+        if self.block == 0:
+            if self.fail:
+                output += "[error]"
+        else:
+            errs, fail = self.parent.get_block_errors(self.block)
+            if errs != 0 or fail:
+                output += "[error]"
+        if self.efuse_class == "keyblock":
+            name = self.parent.blocks[self.block].key_purpose_name
+            if name is not None:
+                output += f"\n  Purpose: {self.parent[name].get()}\n "
+        return output
+
+    def reset(self):
+        # resets a efuse that is prepared for burning
+        bitarray_field = self.convert_to_bitstring(0)
+        block = self.parent.blocks[self.block]
+        wr_bitarray_temp = block.wr_bitarray.copy()
+        position = wr_bitarray_temp.length - (
+            self.word * 32 + self.pos + bitarray_field.len
+        )
+        wr_bitarray_temp.overwrite(bitarray_field, pos=position)
+        block.wr_bitarray = wr_bitarray_temp
