@@ -17,6 +17,7 @@ from .. import util
 from ..base_operations import (
     add_common_commands,
     add_force_write_always,
+    add_show_sensitive_info_option,
     burn_bit,
     burn_block_data,
     burn_efuse,
@@ -42,6 +43,7 @@ def add_commands(subparsers, efuses):
         action="store_true",
     )
     add_force_write_always(p)
+    add_show_sensitive_info_option(p)
     p.add_argument(
         "block",
         help='Key block to burn. "flash_encryption" (block1), '
@@ -89,6 +91,7 @@ def add_commands(subparsers, efuses):
         action="store_true",
     )
     add_force_write_always(burn_key_digest)
+    add_show_sensitive_info_option(burn_key_digest)
 
     p = subparsers.add_parser(
         "set_flash_voltage",
@@ -154,22 +157,27 @@ def set_flash_voltage(esp, efuses, args):
         )
 
     if args.voltage == "OFF":
-        msg = "Disable internal flash voltage regulator (VDD_SDIO). "
-        "SPI flash will need to be powered from an external source.\n"
-        "The following efuse is burned: XPD_SDIO_FORCE.\n"
-        "It is possible to later re-enable the internal regulator (%s) " % (
-            "to 3.3V" if sdio_tieh.get() != 0 else "to 1.8V or 3.3V"
+        print(
+            "Disable internal flash voltage regulator (VDD_SDIO). "
+            "SPI flash will need to be powered from an external source.\n"
+            "The following eFuse is burned: XPD_SDIO_FORCE.\n"
+            "It is possible to later re-enable the internal regulator"
+            f"{'to 3.3V' if sdio_tieh.get() != 0 else 'to 1.8V or 3.3V'}"
+            "by burning an additional eFuse"
         )
-        "by burning an additional efuse"
     elif args.voltage == "1.8V":
-        msg = "Set internal flash voltage regulator (VDD_SDIO) to 1.8V.\n"
-        "The following efuses are burned: XPD_SDIO_FORCE, XPD_SDIO_REG.\n"
-        "It is possible to later increase the voltage to 3.3V (permanently) "
-        "by burning additional efuse XPD_SDIO_TIEH"
+        print(
+            "Set internal flash voltage regulator (VDD_SDIO) to 1.8V.\n"
+            "The following eFuses are burned: XPD_SDIO_FORCE, XPD_SDIO_REG.\n"
+            "It is possible to later increase the voltage to 3.3V (permanently) "
+            "by burning additional eFuse XPD_SDIO_TIEH"
+        )
     elif args.voltage == "3.3V":
-        msg = "Enable internal flash voltage regulator (VDD_SDIO) to 3.3V.\n"
-        "The following efuses are burned: XPD_SDIO_FORCE, XPD_SDIO_REG, XPD_SDIO_TIEH."
-    print(msg)
+        print(
+            "Enable internal flash voltage regulator (VDD_SDIO) to 3.3V.\n"
+            "The following eFuses are burned: XPD_SDIO_FORCE, XPD_SDIO_REG, XPD_SDIO_TIEH."
+        )
+
     sdio_force.save(1)  # Disable GPIO12
     if args.voltage != "OFF":
         sdio_reg.save(1)  # Enable internal regulator
@@ -230,7 +238,14 @@ def burn_key(esp, efuses, args):
         if block_name in ("flash_encryption", "secure_boot_v1"):
             revers_msg = "\tReversing the byte order"
             data = data[::-1]
-        print(" - %s -> [%s]" % (efuse.name, util.hexify(data, " ")))
+        print(" - %s" % (efuse.name), end=" ")
+        print(
+            "-> [{}]".format(
+                util.hexify(data, " ")
+                if args.show_sensitive_info
+                else " ".join(["??"] * len(data))
+            )
+        )
         if revers_msg:
             print(revers_msg)
         if len(data) != num_bytes:
@@ -261,8 +276,10 @@ def burn_key(esp, efuses, args):
             "The key block will left readable and writeable (due to --no-protect-key)"
         )
     else:
-        msg += "The key block will be read and write protected "
-        "(no further changes or readback)"
+        msg += (
+            "The key block will be read and write protected "
+            "(no further changes or readback)"
+        )
     print(msg, "\n")
     if not efuses.burn_all(check_batch_mode=True):
         return
@@ -290,7 +307,14 @@ def burn_key_digest(esp, efuses, args):
             "Digest must be %d bytes (%d bits) of raw binary key data."
             % (len(digest), num_bytes, num_bytes * 8)
         )
-    print(" - %s -> [%s]" % (efuse.name, util.hexify(digest, " ")))
+    print(" - %s" % (efuse.name), end=" ")
+    print(
+        "-> [{}]".format(
+            util.hexify(digest, " ")
+            if args.show_sensitive_info
+            else " ".join(["??"] * len(digest))
+        )
+    )
 
     efuse.save(digest)
     if not args.no_protect_key:
