@@ -102,11 +102,11 @@ class EspEfuses(base_fields.EspEfusesBase):
                 for efuse in self.Fields.BLOCK2_CALIBRATION_EFUSES
             ]
         else:
-            # if self["BLK_VERSION_MINOR"].get() == 1:
-            #     self.efuses += [
-            #         EfuseField.convert(self, efuse)
-            #         for efuse in self.Fields.BLOCK2_CALIBRATION_EFUSES
-            #     ]
+            if self.get_block_version() >= 1:
+                self.efuses += [
+                    EfuseField.convert(self, efuse)
+                    for efuse in self.Fields.BLOCK2_CALIBRATION_EFUSES
+                ]
             self.efuses += [
                 EfuseField.convert(self, efuse) for efuse in self.Fields.CALC
             ]
