@@ -19,10 +19,10 @@ class EmulateEfuseController(EmulateEfuseControllerBase):
 
     def __init__(self, efuse_file=None, debug=False):
         self.Blocks = EfuseDefineBlocks
-        self.Fields = EfuseDefineFields()
+        self.Fields = EfuseDefineFields(None)
         self.REGS = EfuseDefineRegisters
         super(EmulateEfuseController, self).__init__(efuse_file, debug)
-        self.write_reg(self.REGS.EFUSE_STATUS_REG, 1)
+        self.write_reg(self.REGS.EFUSE_CMD_REG, 0)
 
     """ esptool method start >>"""
 
@@ -53,10 +53,10 @@ class EmulateEfuseController(EmulateEfuseControllerBase):
                 self.clean_blocks_wr_regs()
                 self.check_rd_protection_area()
                 self.write_reg(addr, 0)
-                self.write_reg(self.REGS.EFUSE_STATUS_REG, 1)
+                self.write_reg(self.REGS.EFUSE_CMD_REG, 0)
             elif value == self.REGS.EFUSE_READ_CMD:
                 self.write_reg(addr, 0)
-                self.write_reg(self.REGS.EFUSE_STATUS_REG, 1)
+                self.write_reg(self.REGS.EFUSE_CMD_REG, 0)
                 self.save_to_file()
 
     def get_bitlen_of_block(self, blk, wr=False):
