@@ -13,7 +13,7 @@ class ESP8266ROM(ESPLoader):
     CHIP_NAME = "ESP8266"
     IS_STUB = False
 
-    CHIP_DETECT_MAGIC_VALUE = [0xFFF0C101]
+    MAGIC_VALUE = 0xFFF0C101
 
     # OTP ROM addresses
     ESP_OTP_MAC0 = 0x3FF00050
