@@ -10,15 +10,8 @@ class ESP32S3BETA2ROM(ESP32S3ROM):
     CHIP_NAME = "ESP32-S3(beta2)"
     IMAGE_CHIP_ID = 4
 
-    CHIP_DETECT_MAGIC_VALUE = [0xEB004136]
-
     EFUSE_BASE = 0x6001A000  # BLOCK0 read base address
 
-    def get_chip_description(self):
-        major_rev = self.get_major_chip_version()
-        minor_rev = self.get_minor_chip_version()
-        return f"{self.CHIP_NAME} (revision v{major_rev}.{minor_rev})"
-
 
 class ESP32S3BETA2StubLoader(ESP32S3BETA2ROM):
     """Access class for ESP32S3 stub loader, runs on top of ROM.
